#include <windows.h>
#include <gl/glfw.h>
#include <math.h>
#include <stdio.h>
//#include <stdlib.h>
#include "Log.h"

#define PI 3.1415f

#include "Vector.h"
//#include "Texture.h"
//#include "fileselector.h"

Vector Normalize(Vector v)
{
	v.Normalize();
	return v;
}

Vector NormalizeSafe(Vector v)
{
	v.NormalizeSafe();
	return v;
}

bool orthoview=false;

void glVertexV(Vector& vec)
{
	if(orthoview)
		glVertex3f(vec.x, vec.y, 0.0f);
	else
		glVertex3f(vec.x, vec.y, vec.z);
}

float spow(float x, float p)
{
	if(x<0.0f)
		return -pow(-x, p);
	return pow(x, p);
}

Vector RotateY(Vector v, float a)
{
	float cosa=(float)cos(a);
	float sina=(float)sin(a);
	float x=cosa*v.x+sina*v.z;
	float z=-sina*v.x+cosa*v.z;
	v.x=x;
	v.z=z;
	return v;
}

Vector RotateX(Vector v, float a)
{
	float cosa=(float)cos(-a);
	float sina=(float)sin(-a);
	float y=cosa*v.y+sina*v.z;
	float z=-sina*v.y+cosa*v.z;
	v.z=z;
	v.y=y;
	return v;
}

Vector RotateZ(Vector v, float a)
{
	float cosa=(float)cos(-a);
	float sina=(float)sin(-a);
	float x=cosa*v.x+sina*v.y;
	float y=-sina*v.x+cosa*v.y;
	v.x=x;
	v.y=y;
	return v;
}

Vector ProjectPointPlane(Vector p, Vector origin, Vector normal)
{
	p-=origin;
	p-=normal*(normal*p);
	p+=origin;
	return p;
}

#define D2R (3.1415f/180.0f)
#define R2D (180.0f/3.1415f)



struct Vertex
{
	Vector pos;
	Vector normal;
	Vector apos; // deformed position from animation
	Vector anormal; // deformed normal from animation
	int id;
};

struct Triangle
{
	Vertex* verts[3];
	Vector normal;
};

struct Model
{
	Vector size; // maximum length along each axis, from origin
	Vertex* verts;
	int numverts;
	Triangle* tris;
	int numtris;
};

Model player;
Model world;

float* vertex_array;
float* normal_array;
int* index_array;
int array_numtris;

Vector p_pos;
Vector p_dpos;
float p_floordist;
float p_angle;
float p_dangle;

Vector cam_pos;
Vector cam_right;
Vector cam_up;
Vector cam_forward;
Vector cam_pivot;
float camzoom=1.0f;
float rotangle_pan=0.0f;
float rotangle_tilt=0.0f;
float cam_dist;

#include "shader.h"

GLUquadric* gluq;

void InitGL()
{
	glfwSetWindowTitle("DrPetter's LD16 entry");

	glfwSwapInterval(1);

	glfwEnable(GLFW_STICKY_KEYS);

	glViewport(0, 0, 800, 600);
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glClearStencil(0);

	glDisable(GL_DEPTH_TEST);
	glDisable(GL_TEXTURE_2D);

	glDepthFunc(GL_LEQUAL);	
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	
//	glEnable(GL_LINE_SMOOTH);
//	glEnable(GL_POINT_SMOOTH);
//	glEnable(GL_BLEND);
//	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
//	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
//	gluPerspective(45.0f, 800.0f/600.0f, -0.5f, 0.5f);
	glOrtho(0,800,600,0,-1,1);
	glMatrixMode(GL_MODELVIEW);

	gluq=gluNewQuadric();
}

Vector TriNormal(Triangle* t)
{
	return NormalizeSafe((t->verts[1]->pos-t->verts[0]->pos)%(t->verts[2]->pos-t->verts[0]->pos));
}

bool IntersectSphereTriangle(Vector sp, float sr, Triangle* t, Vector& hitpoint)
{
	Vector normal=t->normal;
	Vector d=t->verts[0]->pos-sp;
	float dist=d*normal;
	if(fabs(dist)>sr) // does not intersect triangle plane
		return false;
	Vector pp=sp+normal*dist; // pp is on triangle plane
	// check if pp is inside t (in front of all edges in t)
	Vector eo[3];
	Vector ep[3];
	Vector en[3];
	Vector ed[3];
	bool inside=true;
	for(int i=0;i<3;i++)
	{
//		Edge* e=t->edges[i];
//		eo[i]=e->verts[0]->pos;
//		ep[i]=e->verts[1]->pos;
		eo[i]=t->verts[i]->pos;
		ep[i]=t->verts[(i+1)%3]->pos;
		en[i]=normal%(ep[i]-eo[i]);
		ed[i]=pp-eo[i];
		if(ed[i]*en[i]<0.0f) // "behind" this edge
			inside=false; // can't break out since we're building info that will be used later
	}
	if(inside)
	{
		hitpoint=pp;
		return true;
	}
	// check if closest point on triangle perimeter lies within sphere
	for(int i=0;i<3;i++)
	{
//		Edge* e=t->edges[i];
		Vector et=ep[i]-eo[i];
		float el=et.Length();
		et*=1.0f/el; // normalize
		float f=ed[i]*et;
		if(f<0.0) f=0.0;
		if(f>el) f=el;
		Vector edp=eo[i]+et*f;
		if((edp-sp).Length()<sr)
		{
			hitpoint=edp;
			return true; // it's enough if one edge touches the sphere
		}
	}
	return false;
}


float IntersectRayTriangle(Triangle* t, Vector origin, Vector dir)
{
	Vector normal=t->normal;
	float dy=-dir*normal;
	if(dy<=0.0f) // triangle backfacing
		return -1.0f;
	float dist=(origin-t->verts[0]->pos)*normal;
	dist/=dy;
	Vector pp=origin+dir*dist;
	// check if projected point lies within triangle
	for(int i=0;i<3;i++)
	{
		Vector eo=t->verts[i]->pos;
		Vector ep=t->verts[(i+1)%3]->pos;
		Vector en=normal%(ep-eo);
		Vector ed=pp-eo;
		if(ed*en<0.0f)
			return -1.0f; // "behind" this edge, outside triangle
	}
	return dist;
}

float Raycast(Vector origin, Vector dir) // shoot ray and return nearest hit point on mesh
{
	float closest_dist=100000.0f;
	// for each triangle in facelist:
	for(int i=0;i<world.numtris;i++)
	{
		Triangle* t=&world.tris[i];

		float limit=300.0f;
		Vector d=t->verts[0]->pos-p_pos;
		if(d.x<-limit || d.x>limit)
			continue;
		if(d.y<-limit || d.y>limit)
			continue;
		if(d.z<-limit || d.z>limit)
			continue;

		//  if intersecting ray, note intersection position if closest yet
		float dist=IntersectRayTriangle(t, origin, dir);
		if(dist>0.0f && dist<closest_dist)
		{
			closest_dist=dist;
//			*tri=t; // report which triangle was hit
		}
	}
	if(closest_dist==100000.0f)
		return -1.0f;
	// return closest intersection point
	// * also return triangle pointer?
//	*result=origin+dir*closest_dist;
	return closest_dist;
}

void DrawCircle(Vector pos, float radius, Vector right, Vector up, Vector front)
{
	glBegin(GL_LINES);
	Vector pp;
	for(int i=0;i<=24;i++)
	{
		float a=(float)i/24*3.1415f*2;
		if(i==24)
			a=0.0f;
		if(i>0)
			glVertexV(pp);
		pp=   pos
			+ right*((float)cos(a)*radius)
			+ up*((float)sin(a)*radius)
			- front*0.1f;
//		pp.z=0.0f; // temporarily, for ortho projection
		if(i>0)
			glVertexV(pp);
	}
	glEnd();
}

void DrawCircle(Vector pos, float radius)
{
	DrawCircle(pos, radius, Vector(1,0,0), Vector(0,1,0), Vector(0,0,0));
}

void DrawCubeFace(float x, float y, float z)
{
	Vector normal=Vector(x,y,z);
	Vector tx, ty;
	if(normal.x==0.0f)
	{
		tx=normal%Vector(1,0,0);
		ty=normal%tx;
	}
	else if(normal.y==0.0f)
	{
		tx=normal%Vector(0,1,0);
		ty=normal%tx;
	}
	else
	{
		tx=normal%Vector(0,0,1);
		ty=normal%tx;
	}
	glBegin(GL_QUADS);
	glVertexV(normal+tx-ty);
	glVertexV(normal+tx+ty);
	glVertexV(normal-tx+ty);
	glVertexV(normal-tx-ty);
	glEnd();
}

void DrawCube()
{
	DrawCubeFace(1,0,0);
	DrawCubeFace(-1,0,0);
	DrawCubeFace(0,1,0);
	DrawCubeFace(0,-1,0);
	DrawCubeFace(0,0,1);
	DrawCubeFace(0,0,-1);
}
/*
void DrawBrush(Brush& b, float scale)
{
	// use camera vector for these later
	glPushMatrix();
	glTranslatef(b.pos.x, b.pos.y, b.pos.z);
	gluSphere(gluq, b.radius*scale, 16, 16);
	glPopMatrix();

	return;
}
*/
void SetOrtho(bool yes)
{
	orthoview=yes;

	if(yes)
	{
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
	//	gluPerspective(45.0f, 800.0f/600.0f, -0.5f, 0.5f);
		glOrtho(0,800,600,0,-1,1);
		glMatrixMode(GL_MODELVIEW);
		glDisable(GL_DEPTH_TEST);
	}
	else
	{
		glEnable(GL_DEPTH_TEST);
		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f, 800.0f/600.0f, 5.0f, 3000.0f);
	//	glOrtho(0,800,600,0,-1,1);
		glMatrixMode(GL_MODELVIEW);
	}
}

void DrawQuad(float x, float y, float w, float h)
{
	glBegin(GL_QUADS);
	glTexCoord2f(0,0);
	glVertex3f(x, y, 0);
	glTexCoord2f(0,1);
	glVertex3f(x, y+h, 0);
	glTexCoord2f(1,1);
	glVertex3f(x+w, y+h, 0);
	glTexCoord2f(1,0);
	glVertex3f(x+w, y, 0);
	glEnd();
}

//Texture gui_buttons;

Vector mousepos;
bool gui_mousepressed=false;
bool gui_mouseclick=false;

bool InBox(Vector v, float x, float y, float w, float h)
{
	return (v.x>=x && v.y>=y && v.x<x+w && v.y<y+h);
}

void InitGUI()
{
//	gui_buttons.LoadPNG("data/gui_buttons.png", TEXTURE_NOMIPMAPS);
}

void RenderGUI() // semi-IMGUI, also handles hover detection
{
/*	gui_hover=false;

	SetOrtho(true);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);
	glLoadIdentity();
	glEnable(GL_TEXTURE_2D);
//	glBindTexture(GL_TEXTURE_2D, colormap.texture.getHandle());
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_BLEND);

	// shadows	
	glBindTexture(GL_TEXTURE_2D, gui_shadow.getHandle());
	// buttons
	glColor4f(1,1,1,0.3f);
	DrawQuad(anchor_buttons.x-40.0f, anchor_buttons.y-40.0f, 60.0f+80.0f, 90.0f+80.0f);
	// tool controls
	glColor4f(1,1,1,0.2f);
	DrawQuad(anchor_tool.x, anchor_tool.y, 350.0f, 60.0f);
	// decals
//	glColor4f(1,1,1,0.2f);
//	DrawQuad(anchor_decals.x, anchor_decals.y, 140.0f, 400.0f);


//	glEnd();
	glDisable(GL_BLEND);
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);

	gui_mouseclick=false;*/
}

#define AXIS_X 0
#define AXIS_Y 1
#define AXIS_Z 2

float clamp(float x, float xmin, float xmax)
{
	if(x<xmin) x=xmin;
	if(x>xmax) x=xmax;
	return x;
}

float softabs(float x, float range)
{
	x=fabs(x);
	if(x<range)
//		x=0.0f;
		x=(1.0f-(float)sin((1.0f-x/range)*3.1415f/2))*range;
	return x;
}

void BlendedRotate(Vector& v, Vector& n, int axis, Vector offset, float angle, float blend)
{
	if(blend==0.0f)
		return;
	v-=offset;
	if(axis==AXIS_X)
	{
		v=RotateX(v, angle*blend);
//		n=RotateX(n, angle*blend);
	}
	if(axis==AXIS_Y)
	{
		v=RotateY(v, angle*blend);
//		n=RotateY(n, angle*blend);
	}
	if(axis==AXIS_Z)
	{
		v=RotateZ(v, angle*blend);
//		n=RotateZ(n, angle*blend);
	}
//	n.Normalize();
	v+=offset;
}

void AnimateWalk(Model& model, float t, float total_blend)
{
	float at=t*3.1415f*2;
	float msx=model.size.x;
	float msy=model.size.y;
	float msz=model.size.z;
	for(int i=0;i<model.numverts;i++)
	{
		Vertex& v=model.verts[i];
		float fx=v.pos.x/model.size.x;
		float fy=v.pos.y/model.size.y;
		float fz=v.pos.z/model.size.z;
		v.apos=v.pos;
		v.anormal=v.normal;
		if(total_blend!=0.0f)
		{
			BlendedRotate(v.apos, v.anormal, AXIS_X, Vector(0,0,0), -softabs(sin(at-0.2f), 0.4f)*0.7f+0.2f, clamp(fy+0.25f, 0, 1)-clamp((fy-0.15f)*0.9f, 0, 2.0f));
			float a;
			float w;
			a=spow(sin(at)*0.6f, 0.8f);
			w=clamp(-(fy+0.35f)*3, 0, 1)*clamp(fx*5, 0, 1);
			w=spow(w, 0.5f);
			BlendedRotate(v.apos, v.anormal, AXIS_X, Vector(0,-0.3f*msy,0), a, w);
			a=spow(-sin(at)*0.6f, 0.8f);
			w=clamp(-(fy+0.35f)*3, 0, 1)*clamp(-fx*5, 0, 1);
			w=spow(w, 0.5f);
			BlendedRotate(v.apos, v.anormal, AXIS_X, Vector(0,-0.3f*msy,0), a, w);
			BlendedRotate(v.apos, v.anormal, AXIS_Y, Vector(0,0,0), -sin(at+0.3f)*0.3f, clamp((fy+0.5f)*4, 0, 1)*(0.5f+fx*fx*1.3f));
			BlendedRotate(v.apos, v.anormal, AXIS_Z, Vector(0,-1,0), -sin(at-0.1f)*0.2f, clamp((fy+1.0f), 0, 1));
			v.apos.y=(v.apos.y+model.size.z)*(0.85f+softabs(sin(at+0.3f), 0.4f)*0.15f)-model.size.z;
			v.apos=v.pos+(v.apos-v.pos)*total_blend;
		}
	}
}

void Render()
{
	SetOrtho(false);

	glDepthMask(1);

	glClearColor(0.4f, 1.0f, 0.7f, 0.0f);
	glClear(GL_COLOR_BUFFER_BIT|GL_DEPTH_BUFFER_BIT|GL_STENCIL_BUFFER_BIT);

	glLoadIdentity();
//	glTranslatef(0,0,-500);
//	glScalef(0.1f, 0.1f, 0.1f);
//	glScalef(camzoom, camzoom, camzoom);
//	glTranslatef(-campos.x+400,-campos.y+300,0);
//	glTranslatef(-400,-300,0);
//	glRotatef(30, 1,0,0);
	glTranslatef(0,-30*cam_dist,-250*cam_dist);
	glRotatef(rotangle_tilt, 1,0,0);
	glRotatef(rotangle_pan, 0,1,0);
	glTranslatef(-cam_pos.x,-cam_pos.y,-cam_pos.z);
//	glRotatef(90, 1,0,0);


	EnableShaders(true);

	// draw world
	glColor3f(0,0.1f,0);

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);
//	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glVertexPointer(3, GL_FLOAT, 0, vertex_array);
	glNormalPointer(GL_FLOAT, 0, normal_array);
//	glTexCoordPointer(2, GL_FLOAT, 0, geom->texcoord_array[i]);
	glDrawElements(GL_TRIANGLES, array_numtris*3, GL_UNSIGNED_INT, index_array);

/*	glBegin(GL_TRIANGLES);
	for(int i=0;i<world.numtris;i++)
	{
		Triangle& t=world.tris[i];
		for(int j=0;j<3;j++)
		{
			Vertex* v=t.verts[j];
			Vector* n=&v->normal;
			glNormal3f(n->x, n->y, n->z);
			glVertexV(v->pos);
		}
	}
	glEnd();*/
	EnableShaders(false);

/*
	if(texturemapped)
	{
		glEnable(GL_TEXTURE_2D);
	//	glBindTexture(GL_TEXTURE_2D, texture.getHandle());
	//	glBindTexture(GL_TEXTURE_2D, trimask.texture.getHandle());
		glBindTexture(GL_TEXTURE_2D, colormap.texture.getHandle());
	//	glBindTexture(GL_TEXTURE_2D, decals[0].texture.getHandle());
	
	//	bool textured=false;
	
		glColor3f(1,1,1);
		glBegin(GL_TRIANGLES);
		for(int i=0;i<ALLOC_STORE;i++)
		{
			Triangle* t=&triangle_store[i];
			if(t->active)
			{
				if(t->patch_id>-1)
				{
					if(t->texverts[0]==NULL || t->texverts[1]==NULL || t->texverts[2]==NULL)
						continue;
					Vector& pc=patch_colors[t->patch_id%31];
					for(int j=0;j<3;j++)
					{
						Vector n=t->verts[j]->normal;
						if(n.x<0.0f) n.x=0.0f;
						if(n.y<0.0f) n.y=0.0f;
						Vector c=Vector(0.2f,0.2f,0.2f)+Vector(0.7f,0.3f,0.1f)*n.x+Vector(0.4f,0.4f,0.4f)*n.y;
//						c*=1.1f;
						c=c*0.5f+pc*0.5f;
						glColor3f(c.x, c.y, c.z);
						glTexCoord2f(t->texverts[j]->pos.x, t->texverts[j]->pos.y);
						glVertexV(t->verts[j]->pos);
						
	//					textured=true;
					}
				}
			}
		}
		glEnd();
		
		glDisable(GL_TEXTURE_2D);
	}
*/

	glDepthMask(0);
	// draw shadow
	glPushAttrib(GL_ALL_ATTRIB_BITS);
	glPushMatrix();
	Vector sp=p_pos+Vector(0, -p_floordist, 0);
	glTranslatef(sp.x, sp.y, sp.z);
	float sr=player.size.z;
	glDepthFunc(GL_GEQUAL);
	glEnable(GL_STENCIL_TEST);
	glColorMask(0,0,0,0);
	glStencilOp(GL_KEEP, GL_KEEP, GL_REPLACE);
	// set backfaces
	glStencilFunc(GL_ALWAYS, 0x1, 0x1);
	glCullFace(GL_FRONT);
	gluSphere(gluq, sr, 16, 16);
	// clear frontfaces
	glStencilFunc(GL_ALWAYS, 0x0, 0x1);
	glCullFace(GL_BACK);
	gluSphere(gluq, sr, 16, 16);
	// draw backfaces
	glStencilFunc(GL_EQUAL, 0x1, 0x1);
	glColorMask(1,1,1,1);
	glCullFace(GL_FRONT);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glColor4f(0.0f, 0.0f, 0.0f, 0.2f);
	gluSphere(gluq, sr, 16, 16);
	glDisable(GL_BLEND);
	// restore
	glDisable(GL_STENCIL_TEST);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);	

	glCullFace(GL_BACK);
	glPopMatrix();
	glPopAttrib();
	glDepthMask(1);




	// draw player
	EnableShaders(true);
	glPushMatrix();
	glTranslatef(p_pos.x, p_pos.y, p_pos.z);
	glRotatef(p_angle/3.1415f*180.0f, 0,1,0);
	glColor3f(0.05f,0,0.03f);
	glBegin(GL_TRIANGLES);
	for(int i=0;i<player.numtris;i++)
	{
		Triangle& t=player.tris[i];
		for(int j=0;j<3;j++)
		{
			Vertex* v=t.verts[j];
			Vector* n=&v->anormal;
			glNormal3f(n->x, n->y, n->z);
			glVertexV(v->apos);
		}
	}
	glEnd();
	glPopMatrix();
	EnableShaders(false);



//	RenderGUI();

}

#include "meshfile.h"

#include "playmu/playmu.h"

int main()
{
//	LogStart("log.txt");
	LogDisable();
	LogPrint("hello");
	
	glfwInit();

	if(!glfwOpenWindow(800, 600, 0,0,0,0, 16, 1, GLFW_WINDOW)) // width, height, r, g, b, a, depth, stencil, mode (0 for r,g,b uses desktop settings)
	{
		glfwTerminate();
		return 0;
	}
	InitGL();

	if(!InitShaders())
	{
		LogPrint("### shader init failed!");
//		return 0;
	}

//	timeBeginPeriod(2);

//	InitGUI();


	LoadOBJ(player, "data/mesh.obj", 0.1f);
	LoadOBJ(world, "data/world.obj", 12.0f);

	LogPrint("build arrays");
	vertex_array=new float[world.numverts*3];
	normal_array=new float[world.numverts*3];
	LogPrint("vertex");
	for(int i=0;i<world.numverts*3;i+=3)
	{
		vertex_array[i+0]=world.verts[i/3].pos.x;
		vertex_array[i+1]=world.verts[i/3].pos.y;
		vertex_array[i+2]=world.verts[i/3].pos.z;
		normal_array[i+0]=world.verts[i/3].normal.x;
		normal_array[i+1]=world.verts[i/3].normal.y;
		normal_array[i+2]=world.verts[i/3].normal.z;
	}
	LogPrint("index");
	index_array=new int[world.numtris*3];
	array_numtris=world.numtris;
	for(int i=0;i<world.numtris*3;i+=3)
	{
		index_array[i+0]=world.tris[i/3].verts[0]->id;
		index_array[i+1]=world.tris[i/3].verts[1]->id;
		index_array[i+2]=world.tris[i/3].verts[2]->id;
	}
	LogPrint("arrays done");

	cam_pos=Vector(0,0,0);

	rotangle_pan=160.0f;
	rotangle_tilt=15.0f;

	int step=0;
	
	p_pos=Vector(-40,-270,150);
	p_dpos=Vector(0,0,1.2f);
	p_angle=0.0f;
	p_dangle=0.1f;
	cam_dist=1.0f;

//	unsigned int timestamp=timeGetTime();

	playmu_sound testeffect; // handle for sound effect
	playmu_init();
	playmu_mastervolume(100); // percent of maximum (well, above 100 is possible but not recommended)
	playmu_loadsound(testeffect, "data/sound.wav", 1); // 1 = how many simultaneous playback channels are allowed for this sound
	playmu_loadsong("data/bgm.smu"); // replaces any previously loaded song (can only handle one at a time)
	playmu_playsong(0); // parameter is fade-in time in milliseconds
	playmu_songvolume(25); // percent of original volume
	
	playmu_playsound(testeffect, 100);

	int mx=0;
	int my=0;	
	int mw=glfwGetMouseWheel();
	for(;;)
	{
		if(glfwGetKey(GLFW_KEY_ESC) || !glfwGetWindowParam(GLFW_OPENED))
			break;

		static float walkspeed=0.0f;

//		static int ms=20;
//		static int curt=timestart;
//		curt=timeGetTime()-timestart;
//		static unsigned int pstamp=0;
//		ms+=timeGetTime()-timestart-pstamp;
//		pstamp=timeGetTime()-timestart;

//		while(timeGetTime()-timestamp<30) // enforce 30 fps or such, since logic is run twice... wohoo
//			Sleep(5);

//		timestamp=timeGetTime();
//		LogPrint("loop");
		for(int ttt=0;ttt<1;ttt++)
		{
//			ms+=26; // 1/60 sec
//			timestamp+=20;
//			LogPrint("tick, ms=%i", ms);

			walkspeed-=0.05f;
			if(walkspeed<0.0f)
				walkspeed=0.0f;
			if(walkspeed>1.0f)
				walkspeed=1.0f;
	
			if(glfwGetKey(GLFW_KEY_RIGHT))
			{
				p_dangle-=0.015f;
				walkspeed+=0.1f;
			}
			if(glfwGetKey(GLFW_KEY_LEFT))
			{
				p_dangle+=0.015f;
				walkspeed+=0.1f;
			}
			if(glfwGetKey(GLFW_KEY_UP))
			{
				p_dpos+=Vector((float)sin(p_angle), 0, (float)cos(p_angle))*0.3f;
				walkspeed+=0.1f;
			}
			if(glfwGetKey(GLFW_KEY_DOWN))
			{
				p_dpos-=Vector((float)sin(p_angle), 0, (float)cos(p_angle))*0.3f;
				walkspeed+=0.1f;
			}
	
			step++;
	
			p_dangle*=0.8f;
			if(fabs(p_dangle)<0.01f)
				p_dangle=0.0f;
			p_angle+=p_dangle;
	
			p_dpos+=Vector(0,-0.15f,0);
			p_dpos.x*=0.85f;
			p_dpos.z*=0.85f;
			if(Vector(p_dpos.x, 0, p_dpos.z).Length()<0.2f)
			{
				p_dpos.x=0.0f;
				p_dpos.z=0.0f;
			}
			p_pos+=p_dpos;
	
			float f=Raycast(p_pos, Vector(0,-1,0));
			if(f>0.0f)
				p_floordist=f;
	
			static float dtarget=1.0f;
			float dt=Raycast(p_pos+Vector(0,20,0), -Normalize(Vector((float)sin(p_angle), -0.2f, (float)cos(p_angle))))-5.0f;
			if(dt>0.0f)
				dtarget=dt;
			if(dtarget>260.0f)
				dtarget=260.0f;
			cam_dist+=(dtarget/260.0f-cam_dist)*0.05f;
	
			Vector cam_target=p_pos;
			cam_pos+=(cam_target-cam_pos)*0.05f;
			rotangle_pan+=(-p_angle/3.1415f*180.0f+180.0f-rotangle_pan)*0.02f;
		}

		float bsr=player.size.y*0.95f;
		int bail=0;
		bool hit=true;
		do
		{
			hit=false;
			Vector closest=Vector(0,0,0);
			float closest_dist=bsr;
			float limit=player.size.y*3;
			for(int i=0;i<world.numtris;i++)
			{
				Triangle* t=&world.tris[i];
				Vector d=t->verts[0]->pos-p_pos;
				if(d.x<-limit || d.x>limit)
					continue;
				if(d.y<-limit || d.y>limit)
					continue;
				if(d.z<-limit || d.z>limit)
					continue;
				Vector hitpoint;
				if(IntersectSphereTriangle(p_pos, bsr, t, hitpoint))
				{
					Vector d=hitpoint-p_pos;
					float l=d.Length();
					if(l<closest_dist)
					{
						hit=true;
						closest_dist=l;
						closest=hitpoint;
					}
				}
			}
			if(hit)
			{
				Vector d=closest-p_pos;
				float l=d.Length();
				d.Normalize();
				p_pos-=d*(bsr-l)*1.001f;
				p_dpos-=d*(d*p_dpos);
			}
			bail++;
		} while(hit && bail<2);

		
		AnimateWalk(player, (float)step/40, walkspeed);
		Render();
		
		glfwSwapBuffers();
//		Sleep(10);
	}

//	timeEndPeriod(2);

	delete[] vertex_array;
	delete[] normal_array;
	delete[] index_array;

	delete[] player.verts;
	delete[] player.tris;
	delete[] world.verts;
	delete[] world.tris;

	playmu_close();

	return 0;
}

