bool ReadLine(FILE* file, char* dest)
{
	char c=' ';
	fread(&c, 1, 1, file);
	while(c!='\n' && !feof(file))
	{
		*dest++=c;
		fread(&c, 1, 1, file);
	}
	*dest++='\0';
	if(feof(file))
		return false;
	return true;
}

bool LoadOBJ(Model& model, const char* filename, float scale=1.0f)
{
	LogPrint("load '%s'", filename);
	FILE* file=fopen(filename, "r");
	if(!file)
	{
		LogPrint("file error");
		return false;
	}
	
	model.numverts=0;
	model.numtris=0;

	char line[256];

	LogPrint("scanning for numverts and numtris");

	// read file once to figure out how much data we need to allocate
	while(ReadLine(file, line))
	{
		char cmd[256];
		sscanf(line, "%s ", cmd);
		if(strcmp(cmd, "v")==0) // vertex
			model.numverts++;
		if(strcmp(cmd, "f")==0) // face/triangle
			model.numtris++;
	}

	LogPrint("found %i verts and %i tris", model.numverts, model.numtris);

	fseek(file, 0, 0); // reset file

	model.verts=new Vertex[model.numverts];
	model.tris=new Triangle[model.numtris];

	LogPrint("reading data");

	// read file again to actually get the data
	int vi=0;
	bool vnormals=false;
	int ti=0;
	while(ReadLine(file, line))
	{
		char cmd[256];
		sscanf(line, "%s ", cmd);
		if(strcmp(cmd, "v")==0) // vertex
		{
			float x, y, z;
			sscanf(line, "%s %f %f %f", cmd, &x, &y, &z);
			model.verts[vi].pos=Vector(x, y, z);
			model.verts[vi].id=vi;
			vi++;
		}
		if(strcmp(cmd, "vn")==0) // vertex normal
		{
			if(!vnormals)
			{
				LogPrint("switching to normals");
				vi=0;
				vnormals=true;
			}
			float x, y, z;
			sscanf(line, "%s %f %f %f", cmd, &x, &y, &z);
			model.verts[vi++].normal=Vector(x, y, z);
		}
		if(strcmp(cmd, "f")==0) // face/triangle
		{
			int a, b, c;
			sscanf(line, "%s %i %i %i", cmd, &a, &b, &c);
			model.tris[ti].verts[0]=&model.verts[a-1];
			model.tris[ti].verts[1]=&model.verts[b-1];
			model.tris[ti].verts[2]=&model.verts[c-1];
			model.tris[ti].normal=TriNormal(&model.tris[ti]);
			ti++;
		}
	}

	LogPrint("center mesh and calculate dimensions");
	Vector minpos=Vector(10000,10000,10000);
	Vector maxpos=Vector(-10000,-10000,-10000);
	for(int i=0;i<model.numverts;i++)
	{
		Vertex& v=model.verts[i];
		v.pos*=scale;
		if(v.pos.x<minpos.x) minpos.x=v.pos.x;
		if(v.pos.y<minpos.y) minpos.y=v.pos.y;
		if(v.pos.z<minpos.z) minpos.z=v.pos.z;
		if(v.pos.x>maxpos.x) maxpos.x=v.pos.x;
		if(v.pos.y>maxpos.y) maxpos.y=v.pos.y;
		if(v.pos.z>maxpos.z) maxpos.z=v.pos.z;
	}
	Vector center=(minpos+maxpos)*0.5f;
	for(int i=0;i<model.numverts;i++)
	{
		Vertex& v=model.verts[i];
		v.pos-=center;
		v.pos=RotateY(v.pos, -0.45f);
	}
	model.size=(maxpos-minpos)*0.5f;
	
	LogPrint("success!");
	fclose(file);
	return true;
}

