#include <windows.h>
#include <gl\gl.h>
#include <gl\glu.h>
#include <gl\glext.h>
#include <stdio.h>
#include <math.h>
#include <string.h>

// Multitexturing
//PFNGLMULTITEXCOORD2FARBPROC	glMultiTexCoord2fARB=NULL;
//PFNGLACTIVETEXTUREARBPROC	glActiveTextureARB=NULL;

// Shader functions
PFNGLCREATEPROGRAMOBJECTARBPROC		glCreateProgramObjectARB=NULL;
PFNGLCREATESHADEROBJECTARBPROC		glCreateShaderObjectARB=NULL;
PFNGLGETOBJECTPARAMETERIVARBPROC	glGetObjectParameterivARB=NULL;
PFNGLSHADERSOURCEARBPROC			glShaderSourceARB=NULL;
PFNGLCOMPILESHADERARBPROC			glCompileShaderARB=NULL;
PFNGLLINKPROGRAMARBPROC				glLinkProgramARB=NULL;
PFNGLUSEPROGRAMOBJECTARBPROC		glUseProgramObjectARB=NULL;
PFNGLATTACHOBJECTARBPROC			glAttachObjectARB=NULL;
PFNGLGETINFOLOGARBPROC				glGetInfoLogARB=NULL;
PFNGLGETUNIFORMLOCATIONARBPROC		glGetUniformLocationARB=NULL;

PFNGLUNIFORM3FARBPROC				glUniform3fARB=NULL;
PFNGLUNIFORM1FARBPROC				glUniform1fARB=NULL;

char *ogl_extensionstring;

// Shaders
GLhandleARB vs_pixlight;
GLhandleARB fs_pixlight;
GLhandleARB shader_pixlight;


bool shaders_ok=false;

void EnableShaders(bool enable)
{
	if(!shaders_ok)
		return;
		
	if(!enable)
	{
		glUseProgramObjectARB(0); // disable
		return;
	}
	
	glUseProgramObjectARB(shader_pixlight);

	// Shader parameters
//	Vector lightdir=Vector(1,0,0);
//	Vector observer=Vector(0,0,-500);
	Vector observer=Vector(0,0,500);
	observer=RotateX(observer, -rotangle_tilt*D2R);
	observer=RotateY(observer, -rotangle_pan*D2R);
	observer*=1.0f/camzoom;
	Vector light1dir=Vector(0,1,0);
	light1dir.NormalizeSafe();
//	lightdir=RotateX(lightdir, -rotangle_tilt*D2R);
//	lightdir=RotateY(lightdir, -rotangle_pan*D2R);
	Vector light2dir=Vector(0,-1,0);
//	Vector light2dir=p_pos+Vector(0, player.size.y*1.5f, 0);
//	Vector light2dir=p_pos+Vector(0, 150.0f, 0);
	light2dir.NormalizeSafe();

	char varname[100];
	GLint loc;
	strcpy(varname, "light1_dir");
	loc=glGetUniformLocationARB(shader_pixlight, varname);
    if(loc==-1) // couldn't find variable
    	LogPrint("Unfound variable \"%s\"", varname);
    glUniform3fARB(loc, light1dir.x, light1dir.y, light1dir.z);

/*	strcpy(varname, "light2_dir");
	loc=glGetUniformLocationARB(shader_pixlight, varname);
    if(loc==-1) // couldn't find variable
    	LogPrint("Unfound variable \"%s\"", varname);
    glUniform3fARB(loc, light2dir.x, light2dir.y, light2dir.z);
*/
	strcpy(varname, "observer_pos");
	loc=glGetUniformLocationARB(shader_pixlight, varname);
    if(loc==-1) // couldn't find variable
    	LogPrint("Unfound variable \"%s\"", varname);
    glUniform3fARB(loc, observer.x, observer.y, observer.z);

/*	strcpy(varname, "object_invscale");
	loc=glGetUniformLocationARB(shader_pixlight, varname);
    if(loc==-1) // couldn't find variable
    	LogPrint("Unfound variable \"%s\"", varname);
    glUniform1fARB(loc, 1.0f/100.0f);*/
}

void ShowLog(GLhandleARB shader_object)
{
	int blen=0;
	int slen=0;	
	char *compiler_log;
	
	if(shader_object==0)
		return; // not a valid program object
		
	glGetObjectParameterivARB(shader_object, GL_OBJECT_INFO_LOG_LENGTH_ARB , &blen);
	if(blen>1)
	{
		compiler_log=(char*)malloc(blen);
		glGetInfoLogARB(shader_object, blen, &slen, compiler_log);
		if(compiler_log!=0)
			LogPrint("compiler_log: %s", compiler_log);
	}
	free(compiler_log); 
}

char* ReadString(char *filename)
{
	int filesize, i;
	char *string;
	FILE *file;
	file=fopen(filename, "rb");
	if(!file)
		return NULL;
	fseek(file, 0, SEEK_END);
	filesize=ftell(file);
	fseek(file, 0, 0);
	string=(char*)malloc(filesize+1);
	fread(string, filesize, 1, file);
	string[filesize]='\0';
	fclose(file);
	return string;
}

bool LoadShader(char *filename, GLhandleARB &shader_object)
{
	char *shader_source=ReadString(filename);
	if(!shader_source)
		return false;

	GLint length=(GLint)strlen(shader_source);

//	LogPrint("Compiling program \"%s\" with length %i:\n%s", filename, length, shader_source);

	int success=0;
	glShaderSourceARB(shader_object, 1, (const GLcharARB**)&shader_source, &length);
	free(shader_source);
	glCompileShaderARB(shader_object); 
	glGetObjectParameterivARB(shader_object, GL_OBJECT_COMPILE_STATUS_ARB, &success);
	if(!success)
	{
		LogPrint("*** Shader compilation error");
		ShowLog(shader_object);
		return false;
	}

	return true;
}



bool LoadVertexShader(char *filename, GLhandleARB &shader_object)
{
	shader_object=glCreateShaderObjectARB(GL_VERTEX_SHADER_ARB);
	return LoadShader(filename, shader_object);
}

bool LoadFragmentShader(char *filename, GLhandleARB &shader_object)
{
	shader_object=glCreateShaderObjectARB(GL_FRAGMENT_SHADER_ARB);
	return LoadShader(filename, shader_object);
}



void AddShader(GLhandleARB &program_object, GLhandleARB shader_object)
{
	glAttachObjectARB(program_object, shader_object);
}

bool LinkShader(GLhandleARB &program_object)
{
	int success;
	glLinkProgramARB(program_object);
	glGetObjectParameterivARB(program_object, GL_OBJECT_LINK_STATUS_ARB, &success);
	if(!success)
		return false;
	return true;
}

bool ExtensionSupported(char *name)
{
	if(strstr(ogl_extensionstring, name))
		return true;
	return false;
}

bool InitGlsl()
{
	if(!ExtensionSupported("GL_ARB_vertex_shader"))
		return false;
	if(!ExtensionSupported("GL_ARB_fragment_shader"))
		return false;
	if(!ExtensionSupported("GL_ARB_shader_object"))
		return false;
	if(!ExtensionSupported("GL_ARB_shading_language_100"))
		return false;
		
	return true;
}

bool InitShaders()
{
/*	glActiveTextureARB=(PFNGLACTIVETEXTUREARBPROC)wglGetProcAddress("glActiveTextureARB");
	glMultiTexCoord2fARB=(PFNGLMULTITEXCOORD2FARBPROC)wglGetProcAddress("glMultiTexCoord2fARB");

	if(!glActiveTextureARB || !glMultiTexCoord2fARB)
	{
		LogPrint("Failed to init multitexturing");
		return false;
	}
*/
	glCreateProgramObjectARB=(PFNGLCREATEPROGRAMOBJECTARBPROC)wglGetProcAddress("glCreateProgramObjectARB");
	glCreateShaderObjectARB=(PFNGLCREATESHADEROBJECTARBPROC)wglGetProcAddress("glCreateShaderObjectARB");
	glGetObjectParameterivARB=(PFNGLGETOBJECTPARAMETERIVARBPROC)wglGetProcAddress("glGetObjectParameterivARB");
	glShaderSourceARB=(PFNGLSHADERSOURCEARBPROC)wglGetProcAddress("glShaderSourceARB");
	glCompileShaderARB=(PFNGLCOMPILESHADERARBPROC)wglGetProcAddress("glCompileShaderARB");
	glLinkProgramARB=(PFNGLLINKPROGRAMARBPROC)wglGetProcAddress("glLinkProgramARB");
	glUseProgramObjectARB=(PFNGLUSEPROGRAMOBJECTARBPROC)wglGetProcAddress("glUseProgramObjectARB");
	glAttachObjectARB=(PFNGLATTACHOBJECTARBPROC)wglGetProcAddress("glAttachObjectARB");
	glGetInfoLogARB=(PFNGLGETINFOLOGARBPROC)wglGetProcAddress("glGetInfoLogARB");
	glGetUniformLocationARB=(PFNGLGETUNIFORMLOCATIONARBPROC)wglGetProcAddress("glGetUniformLocationARB");

	glUniform3fARB=(PFNGLUNIFORM3FARBPROC)wglGetProcAddress("glUniform3fARB");
	glUniform1fARB=(PFNGLUNIFORM1FARBPROC)wglGetProcAddress("glUniform1fARB");

	if( !glCreateProgramObjectARB ||
		!glCreateShaderObjectARB ||
		!glGetObjectParameterivARB ||
		!glShaderSourceARB ||
		!glCompileShaderARB ||
		!glLinkProgramARB ||
		!glUseProgramObjectARB ||
		!glAttachObjectARB ||
		!glGetInfoLogARB ||
		!glGetUniformLocationARB ||
		!glUniform3fARB ||
		!glUniform1fARB)
	{
		LogPrint("Failed to init shader functions");
		return false;
	}

	ogl_extensionstring=(char*)glGetString(GL_EXTENSIONS);
	
	LogPrintf("Initializing GLSL...");
	if(!InitGlsl())
		return false;
	LogPrintf(" Success\n");
	LogPrintf("Loading shaders...");
	if(!LoadVertexShader("data/clay_vs.gls", vs_pixlight))
		return false;
	if(!LoadFragmentShader("data/clay_fs.gls", fs_pixlight))
		return false;
	LogPrintf(" Success\n");
	LogPrintf("Linking shaders...");
	shader_pixlight=glCreateProgramObjectARB();
	AddShader(shader_pixlight, vs_pixlight);
	AddShader(shader_pixlight, fs_pixlight);
	if(!LinkShader(shader_pixlight))
		return false;
	LogPrintf(" Success\n");
	LogPrintf("Shaders are go!\n");

	shaders_ok=true;
	
	return true;
}

